@@ -12,7 +12,7 @@ from msdsl.expr.expr import ModelExpr, wrap_constant, ModelOperator, Constant, A
 from msdsl.expr.table import Table, RealTable, UIntTable, SIntTable
 from msdsl.expr.format import UIntFormat, SIntFormat, RealFormat, IntFormat
 from msdsl.expr.signals import Signal, AnalogSignal, DigitalSignal, AnalogInput, AnalogOutput, DigitalOutput, \
-    DigitalInput
+    DigitalInput, DigitalParameter, RealParameter
 from msdsl.generator.tree_op import tree_op
 from msdsl.generator.svreal import compile_range_expr, compile_width_expr, compile_exponent_expr
 from msdsl.expr.analyze import signal_names, signal_name
@@ -134,7 +134,8 @@ class VerilogGenerator(CodeGenerator):
         else:
             raise Exception(f'Input and output formats do not match: {input_.name} with {input_.format_} vs. {output.name} with {output.format_}')
 
-    def make_mem(self, next_: Signal, curr: Signal, init: Number=0, clk: Signal=None, rst: Signal=None, ce: Signal = None):
+    def make_mem(self, next_: Signal, curr: Signal, init: Union[Number, DigitalParameter, RealParameter]=0,
+                 clk: Signal=None, rst: Signal=None, ce: Signal = None):
         # set defaults
         clk_name = clk.name if clk is not None else '`CLK_MSDSL'
 
@@ -150,18 +151,45 @@ class VerilogGenerator(CodeGenerator):
 
         # create memory for real number
         if isinstance(next_.format_, RealFormat) and isinstance(curr.format_, RealFormat):
-            self.macro_call('DFF_INTO_REAL', next_.name, curr.name, rst_name, clk_name, ce_name, str(init))
+            # determine string expression for the initial value
+            if isinstance(init, Number):
+                init_str = str(init)
+            elif isinstance(init, RealParameter):
+                init_str = init.param_name
+            elif isinstance(init, str):
+                init_str = init
+            else:
+                raise Exception(f'Could not determine string representation for initial value {init}')
+
+            # call the macro to create the memory
+            self.macro_call('DFF_INTO_REAL', next_.name, curr.name, rst_name, clk_name, ce_name, init_str)
         # create memory for integer
         elif (isinstance(next_.format_, SIntFormat) and isinstance(curr.format_, SIntFormat)) or \
              (isinstance(next_.format_, UIntFormat) and isinstance(curr.format_, UIntFormat)):
-            # check that initial value is valid
-            assert (curr.format_.min_val <= init <= curr.format_.max_val), \
-                f'Initial value {init} does not fit in the range [{curr.format_.min_val}, {curr.format_.max_val}] of signal {curr.name}.'
+            # check that initial value is valid, if it's number. range checking for DigitalParameters is not
+            # yet implemented
+            if isinstance(init, Number):
+                assert (curr.format_.min_val <= init <= curr.format_.max_val), \
+                    f'Initial value {init} does not fit in the range [{curr.format_.min_val}, {curr.format_.max_val}] of signal {curr.name}.'
+            else:
+                print(f'Warning: could not validate range of init value {init}')
 
-            # check that the widths match
+            # check that the widths of next_ and curr match
             assert next_.format_.width == curr.format_.width, f'The widths of {next_.name} ({next_.format_.width}) does not match the width of {curr.name} ({curr.format_.width}).'
 
-            self.macro_call('MEM_INTO_DIGITAL', next_.name, curr.name, ce_name, clk_name, rst_name, str(init), str(next_.format_.width))
+            # determine string expression for the initial value
+            if isinstance(init, Number):
+                init_str = str(init)
+            elif isinstance(init, DigitalParameter):
+                init_str = init.name
+            elif isinstance(init, str):
+                init_str = init
+            else:
+                raise Exception(f'Could not determine string representation for initial value {init}')
+
+            # call the macro to create the memory
+            self.macro_call('MEM_INTO_DIGITAL', next_.name, curr.name, ce_name, clk_name, rst_name, init_str,
+                            str(next_.format_.width))
         else:
             raise Exception(f'Next and current formats do not match: {next_.name} with {next_.format_} vs. {curr.name} with {curr.format_}')
 
@@ -184,7 +212,25 @@ class VerilogGenerator(CodeGenerator):
         else:
             raise Exception(f'Unknown table type: {type(table)}')
 
-    def start_module(self, name: str, ios: List[Signal], real_params: List):
+    def make_sync_ram(self, signal: AnalogSignal, format_: RealFormat, addr: DigitalSignal,
+                      clk: DigitalSignal=None, ce: DigitalSignal=None, we: DigitalSignal=None,
+                      din: DigitalSignal=None):
+        # set defaults
+        din_name = din.name if din is not None else "'0"
+        clk_name = clk.name if clk is not None else "`CLK_MSDSL"
+        ce_name = ce.name if ce is not None else "1'b1"
+        we_name = we.name if we is not None else "1'b0"
+
+        # call the SVREAL macro
+        self.macro_call('SYNC_RAM_INTO_REAL', addr.name, din_name, signal.name,
+                        clk_name, ce_name, we_name, addr.format_.width, format_.width,
+                        format_.exponent)
+
+    def start_module(self, name: str, ios: List[Signal], real_params: List, digital_params: List=None):
+        # set defaults
+        if digital_params is None:
+            digital_params = []
+
         # clear default nettype to make debugging easier
         self.default_nettype('none')
         self.writeln()
@@ -195,6 +241,14 @@ class VerilogGenerator(CodeGenerator):
         # parameters
         parameters = []
         parameters += [f'parameter real {real_param.param_name}={real_param.default}' for real_param in real_params]
+        for dig_param in digital_params:
+            dig_param_str = 'parameter '
+            if dig_param.signed:
+                dig_param_str += 'signed '
+            if dig_param.width > 1:
+                dig_param_str += f'[{dig_param.width-1}:0] '
+            dig_param_str += f'{dig_param.name}={dig_param.default}'
+            parameters += [dig_param_str]
         parameters += [self.real_param_str(io) for io in ios if isinstance(io.format_, RealFormat)]
         if len(parameters) > 0:
             self.write(' #')
