@@ -6,15 +6,18 @@ from copy import deepcopy
 from pathlib import Path
 
 from math import ceil, log2
+from scipy.stats import truncnorm
+import random
 
-from msdsl.assignment import ThisCycleAssignment, NextCycleAssignment, BindingAssignment, SyncRomAssignment, Assignment
+from msdsl.assignment import (ThisCycleAssignment, NextCycleAssignment, BindingAssignment,
+                              SyncRomAssignment, Assignment, SyncRamAssignment)
 from msdsl.expr.analyze import signal_names
 from msdsl.eqn.cases import address_to_settings
 from msdsl.eqn.eqn_sys import EqnSys
 from msdsl.expr.expr import (ModelExpr, array, concatenate, sum_op, wrap_constant, min_op, clamp_op,
                              to_sint, to_uint)
 from msdsl.expr.signals import (AnalogInput, AnalogOutput, DigitalInput, DigitalOutput, Signal, AnalogSignal,
-                   AnalogState, DigitalState, RealParameter, DigitalSignal)
+                   AnalogState, DigitalState, RealParameter, DigitalSignal, DigitalParameter)
 from msdsl.generator.generator import CodeGenerator
 from msdsl.util import Namer
 from msdsl.eqn.lds import LdsCollection
@@ -22,7 +25,8 @@ from msdsl.expr.format import RealFormat, IntFormat, is_signed
 from msdsl.expr.extras import if_
 from msdsl.circuit import Circuit
 from msdsl.expr.table import Table, RealTable, SIntTable, UIntTable
-from msdsl.function import Function
+from msdsl.function import GeneralFunction, Function, PlaceholderFunction
+from msdsl.lfsr import LFSR
 
 from scipy.signal import cont2discrete
 
@@ -44,9 +48,11 @@ class MixedSignalModel:
         self.probes = []
         self.circuits = []
         self.real_params = []
+        self.digital_params = []
         self.lookup_tables = []
         self.namers = {}
         self.namer = Namer()
+        self._inv_cdf_func = None
 
         # add ios
         for io in ios:
@@ -120,29 +126,31 @@ class MixedSignalModel:
         """
         return self.add_signal(AnalogState(name=name, range_=range_, width=width, exponent=exponent, init=init))
 
-    def add_digital_signal(self, name, width=1, signed=False):
+    def add_digital_signal(self, name, width=1, signed=False, min_val=None, max_val=None):
         """
         Allows for a digital signal to be declared ahead of time with a given format.  In general, it is preferable
         to use bind_name for this, rather than calling add_digital_signal followed by set_this_cycle.
         """
-        return self.add_signal(DigitalSignal(name=name, width=width, signed=signed))
+        return self.add_signal(DigitalSignal(name=name, width=width, signed=signed, min_val=min_val, max_val=max_val))
 
-    def add_digital_input(self, name, width=1, signed=False):
-        return self.add_signal(DigitalInput(name=name, width=width, signed=signed))
+    def add_digital_input(self, name, width=1, signed=False, min_val=None, max_val=None):
+        return self.add_signal(DigitalInput(name=name, width=width, signed=signed, min_val=min_val, max_val=max_val))
 
-    def add_digital_output(self, name, width=1, signed=False, init=0):
+    def add_digital_output(self, name, width=1, signed=False, init=0, min_val=None, max_val=None):
         """
         Note that the initial value will only be used if the digital output has state.
         """
-        return self.add_signal(DigitalOutput(name=name, width=width, signed=signed, init=init))
+        return self.add_signal(DigitalOutput(name=name, width=width, signed=signed, init=init,
+                                             min_val=min_val, max_val=max_val))
 
-    def add_digital_state(self, name, width=1, signed=False, init=0):
+    def add_digital_state(self, name, width=1, signed=False, init=0, min_val=None, max_val=None):
         """
 
         :param init:    The initial value of the digital signal.  This is the reset value of the state variable when
                         `RST_MSDSL is asserted (synchronous reset).
         """
-        return self.add_signal(DigitalState(name=name, width=width, signed=signed, init=init))
+        return self.add_signal(DigitalState(name=name, width=width, signed=signed, init=init,
+                                            min_val=min_val, max_val=max_val))
 
     # signal access functions
 
@@ -185,15 +193,18 @@ class MixedSignalModel:
 
         return self.set_this_cycle(signal=signal, expr=expr)
 
-    def bind_name(self, signal: Union[Signal, str], expr: ModelExpr):
+    def bind_name(self, signal: Union[Signal, str], expr: ModelExpr, range_=None,
+                  width=None, exponent=None):
         """
         TODO: consider deprecating.
         Alias for set_this_cycle.
         """
 
-        return self.set_this_cycle(signal=signal, expr=expr)
+        return self.set_this_cycle(signal=signal, expr=expr, range_=range_,
+                                   width=width, exponent=exponent)
 
-    def set_this_cycle(self, signal: Union[Signal, str], expr: ModelExpr):
+    def set_this_cycle(self, signal: Union[Signal, str], expr: ModelExpr, range_=None,
+                       width=None, exponent=None):
         """
         The behavior of this function is essentially a blocking assignment (in Verilog nomenclature). The provided
         expression is continuously written to the provided signal.
@@ -205,7 +216,15 @@ class MixedSignalModel:
 
         if isinstance(signal, str):
             expr = wrap_constant(expr)
-            signal = self.add_signal(Signal(name=signal, format_=expr.format_))
+            format_ = deepcopy(expr.format_)
+            if isinstance(format_, RealFormat):
+                if range_ is not None:
+                    format_.range_ = range_
+                if exponent is not None:
+                    format_.exponent = exponent
+                if width is not None:
+                    format_.width = width
+            signal = self.add_signal(Signal(name=signal, format_=format_))
             assignment_cls = BindingAssignment
         elif isinstance(signal, Signal):
             assignment_cls = ThisCycleAssignment
@@ -236,6 +255,98 @@ class MixedSignalModel:
 
         return self.add_assignment(NextCycleAssignment(signal=signal, expr=expr, clk=clk, rst=rst, ce=ce))
 
+    def lfsr_signal(self, width: int, clk=None, rst=None, ce=None, name=None, init=None):
+        """
+        Returns a digital signal that is updated according to an LFSR equation.  The values returned will range
+        between 0 and (1<<width)-2.  All of these values will appear once in a pseudo-random order before repeating,
+        so the period is (1<<width)-1.  For example, when width=3, the following sequence loops: [0, 1, 3, 6, 5, 2, 4],
+        with the starting point is determined by the value of "init".
+
+        :param width:   Width of the LFSR state signal that will be returned.  The LFSR equation is determined
+                        based on the width, such that the LFSR period is as long as possible.  In particular,
+                        the period will be of length (1<<width)-1.
+        :param clk:     Optional input.  Will use `CLK_MSDSL by default.
+        :param rst:     Optional input for synchronous reset.  Will use `RST_MSDSL by default.
+        :param ce:      Optional input for clock enable.  Will use "1" (i.e., always enabled) by default.
+        :param name:    Naming prefix to use for the LFSR signal
+        :param init:    Initial value
+        :return:
+        """
+
+        # set defaults
+        if name is None:
+            name = self.get_next_name(f'lfsr_state_')
+        if init is None:
+            init = random.randint(0, (1<<width)-2)
+
+        # validate input
+        assert init != (1<<width)-1, "Invalid LFSR initialization -- will remain stuck at all 1's"
+
+        # create the LFSR state as a digital signal
+        lfsr_state = self.add_digital_state(name, width=width, init=init)
+
+        # update the LFSR signal according to the equation for the given LFSR width
+        lfsr = LFSR(width)
+        self.set_next_cycle(lfsr_state, lfsr.next_state(lfsr_state), clk=clk, rst=rst, ce=ce)
+
+        # return the signal representing the LFSR state
+        return lfsr_state
+
+    def uniform_signal(self, min_val=0.0, max_val=1.0, clk=None, rst=None, ce=None, lfsr_name=None,
+                       lfsr_width=32, lfsr_init=None, uniform_name=None):
+        # set defaults
+        if uniform_name is None:
+            uniform_name = self.get_next_name(f'uniform_var_')
+        if lfsr_name is None:
+            lfsr_name = uniform_name + '_lfsr_state'
+
+        # create the LFSR signal
+        lfsr_signal = self.lfsr_signal(width=lfsr_width, clk=clk, rst=rst, ce=ce, name=lfsr_name,
+                                       init=lfsr_init)
+
+        # bind the uniform variable
+        scale_factor = (max_val-min_val)/((1<<lfsr_width)-1)
+        uniform_expr = (lfsr_signal * scale_factor) + min_val
+
+        return self.bind_name(uniform_name, uniform_expr)
+
+    def arbitrary_noise(self, inv_cdf_func, clk=None, rst=None, ce=None, lfsr_width=32,
+                        lfsr_init=None, lfsr_name=None, uniform_name=None, noise_name=None):
+        # set defaults
+        if noise_name is None:
+            noise_name = self.get_next_name(f'arb_noise_')
+        if uniform_name is None:
+            uniform_name = noise_name + 'unf_var'
+        if lfsr_name is None:
+            lfsr_name = noise_name + '_lfsr'
+
+        unf_sig =self.uniform_signal(clk=clk, rst=rst, ce=ce, lfsr_name=lfsr_name, lfsr_width=lfsr_width,
+                                     lfsr_init=lfsr_init, uniform_name=uniform_name)
+
+        return self.set_from_sync_func(signal=noise_name, func=inv_cdf_func, in_=unf_sig,
+                                       clk=clk, ce=ce, rst=rst)
+
+    def set_gaussian_noise(self, signal: Union[Signal, str], mean=None, std=None, clk=None, rst=None,
+                           ce=None, lfsr_width=32, lfsr_init=None, lfsr_name=None, uniform_name=None,
+                           noise_name=None, func_order=1, func_numel=512, num_sigma=6):
+        # create the inverse CDF function if needed
+        if self._inv_cdf_func is None:
+            inv_cdf = lambda x: truncnorm.ppf(x, -num_sigma, +num_sigma)
+            self._inv_cdf_func = self.make_function(inv_cdf, domain=[0.0, 1.0], order=func_order, numel=func_numel)
+
+        # generate a standard gaussian noise source
+        expr = self.arbitrary_noise(self._inv_cdf_func, clk=clk, rst=rst, ce=ce, lfsr_width=lfsr_width,
+                                    lfsr_init=lfsr_init, lfsr_name=lfsr_name, uniform_name=uniform_name,
+                                    noise_name=noise_name)
+
+        # scale and shift based on the standard deviation and mean
+        if (std is not None) and (std != 1):
+            expr = expr * std
+        if (mean is not None) and (mean != 0):
+            expr = expr + mean
+
+        return self.set_this_cycle(signal, expr)
+
     def set_from_sync_rom(self, signal: Union[Signal, str], table: Table, addr: ModelExpr, clk=None, ce=None):
         """
         The behavior of this operation is a lookup from a synchronous ROM (which should synthesize
@@ -259,14 +370,28 @@ class MixedSignalModel:
         return self.add_assignment(SyncRomAssignment(signal=signal, table=table, addr=addr,
                                                      clk=clk, ce=ce, should_bind=should_bind))
 
-    def set_from_sync_func(self, signal: Union[Signal, str], func: Function, in_: ModelExpr,
-                           clk=None, ce=None, rst=None):
+    def set_from_sync_ram(self, signal: Union[AnalogSignal, str], format_: RealFormat,
+                          addr: ModelExpr, clk=None, ce=None, we=None, din=None):
+        # bind the signal if necessary
+        should_bind = False
+        if isinstance(signal, str):
+            signal = self.add_signal(Signal(name=signal, format_=format_))
+            should_bind = True
+
+        # return the assignment
+        return self.add_assignment(SyncRamAssignment(signal=signal, format_=format_, addr=addr, clk=clk,
+                                                     ce=ce, we=we, din=din, should_bind=should_bind))
+
+    def set_from_sync_func(self, signal: Union[Signal, str], func: GeneralFunction, in_: ModelExpr,
+                           clk=None, ce=None, rst=None, we=None, wdata=None, waddr=None):
         """
         The behavior of this operation is a an evaluation of a Function.
         There is a one-cycle delay in this operation.
 
         :param signal:   Signal object being assigned
-        :param function: Function object
+        :param function: If a Function is used, ROMs are used for the lookup tables.  Otherwise,
+                         if a PlaceholderFunction is used, then RAMs are used, with the expectation
+                         that the user will set the contents of those RAMs before using the function.
         :param in_:      Real-number input of the function
         :param clk:      Optional input.  Will use `CLK_MSDSL by default.
         :param ce:       Optional input for clock enable.  Will use "1" (i.e., always enabled) by default.
@@ -278,11 +403,10 @@ class MixedSignalModel:
         addr_real_expr = (in_ - func.domain[0]) * ((func.numel - 1) / (func.domain[1] - func.domain[0]))
         if func.clamp:
             addr_real_expr = clamp_op(addr_real_expr, 0.0, func.numel-1.0)
-        addr_real = self.set_this_cycle(self.get_next_name(f'{func.name}_addr_real_'), addr_real_expr)
+        addr_real = self.set_this_cycle(self.get_next_name(f'{func.name}_addr_real_'), addr_real_expr,
+                                        range_=func.numel-1.0)
 
         # convert address to signed integer
-        # TODO: avoid need to re-clamp expression; this is a limitation
-        # of the way ranges for real numbers are represented
         addr_sint_expr = to_sint(addr_real, width=func.addr_bits+1)
         if func.clamp:
             addr_sint_expr = clamp_op(addr_sint_expr, 0, func.numel - 1)
@@ -292,14 +416,26 @@ class MixedSignalModel:
         addr_uint_expr = to_uint(addr_sint, width=func.addr_bits)
         addr_uint = self.set_this_cycle(self.get_next_name(f'{func.name}_addr_uint_'), addr_uint_expr)
 
+        # if needed, mux address and
+        if isinstance(func, PlaceholderFunction):
+            addr_mux_expr = if_(we, waddr, addr_uint)
+            addr_mux = self.set_this_cycle(self.get_next_name(f'{func.name}_addr_mux_'), addr_mux_expr)
+
         # calculate fractional address
         addr_frac_expr = addr_real - addr_sint
-        addr_frac = self.set_this_cycle(self.get_next_name(f'{func.name}_addr_frac_'), addr_frac_expr)
+        addr_frac = self.set_this_cycle(self.get_next_name(f'{func.name}_addr_frac_'), addr_frac_expr,
+                                        range_=1.01)
 
         # look up coefficient values
         coeffs = [self.get_next_name(f'{func.name}_coeff_{k}_') for k in range(func.order+1)]
-        for coeff, table in zip(coeffs, func.tables):
-            self.set_from_sync_rom(signal=coeff, table=table, addr=addr_uint, clk=clk, ce=ce)
+        for k, coeff in enumerate(coeffs):
+            if isinstance(func, PlaceholderFunction):
+                self.set_from_sync_ram(signal=coeff, format_=func.formats[k], addr=addr_mux,
+                                       clk=clk, ce=ce, we=we, din=wdata[k])
+            elif isinstance(func, Function):
+                self.set_from_sync_rom(signal=coeff, table=func.tables[k], addr=addr_uint, clk=clk, ce=ce)
+            else:
+                raise Exception(f'Unknown function type: {func}')
 
         # compute higher-order products of terms
         prods_imm = [self.get_next_name(f'{func.name}_prod_imm_{k}_') for k in range(func.order)]
@@ -386,7 +522,6 @@ class MixedSignalModel:
 
         # create the table, register it, and return it
         function = Function(func=func, domain=domain, name=name, dir=dir, **kwargs)
-        function.create_tables()
 
         # append values
         self.lookup_tables.extend(function.tables)
@@ -408,7 +543,7 @@ class MixedSignalModel:
 
     # parameter functions
 
-    def add_real_param(self, name: str, default: Number):
+    def add_real_param(self, name: str, default: Number=0):
         """
         Equivalent to a real parameter in a Verilog module definition.  Allows the user to generate a single
         SystemVerilog model that can be used for various purposes.
@@ -425,6 +560,24 @@ class MixedSignalModel:
 
         return param
 
+    def add_digital_param(self, name: str, width=1, signed=False, default=0, min_val=None, max_val=None):
+        """
+        Equivalent to a signed/unsigned parameter in a Verilog module definition.  Allows the user to generate a single
+        SystemVerilog model that can be used for various purposes.
+
+        :param name:        Name of the parameter.
+        :param default:     Real number default for the parameter.  A ModelExpr is not allowed here.
+        :return:            Returns a signal representing the parameter that can be used in subsequent expressions.
+        """
+
+        param = DigitalParameter(name=name, width=width, signed=signed, default=default, min_val=min_val,
+                                 max_val=max_val)
+        self.digital_params.append(param)
+
+        setattr(self, name, param)
+
+        return param
+
     # signal probe functions
 
     def add_probe(self, signal: Signal):
@@ -724,7 +877,8 @@ class MixedSignalModel:
                 internals.append(signal)
 
         # start module
-        gen.start_module(name=self.module_name, ios=ios, real_params=self.real_params)
+        gen.start_module(name=self.module_name, ios=ios, real_params=self.real_params,
+                         digital_params=self.digital_params)
 
         # declare the internal variables
         if len(internals) > 0:
@@ -752,6 +906,10 @@ class MixedSignalModel:
             elif isinstance(assignment, SyncRomAssignment):
                 gen.make_sync_rom(signal=assignment.signal, table=assignment.table,
                                   addr=result, clk=assignment.clk, ce=assignment.ce)
+            elif isinstance(assignment, SyncRamAssignment):
+                gen.make_sync_ram(signal=assignment.signal, format_=assignment.format_,
+                                  addr=result, clk=assignment.clk, ce=assignment.ce,
+                                  we=assignment.we, din=assignment.din)
             else:
                 raise Exception('Invalid assignment type.')
 
